package org.processmining.humbertogarciacaballero.models.gui.render;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.RectangularShape;
import java.awt.geom.RoundRectangle2D;

import org.processmining.humbertogarciacaballero.models.util.Utils;

import prefuse.Constants;
import prefuse.render.LabelRenderer;
import prefuse.util.ColorLib;
import prefuse.util.FontLib;
import prefuse.util.GraphicsLib;
import prefuse.util.StringLib;
import prefuse.visual.DecoratorItem;
import prefuse.visual.VisualItem;

public class CustomLabelRenderer extends LabelRenderer {

	private AffineTransform m_transform = new AffineTransform();

	public CustomLabelRenderer() {
		this("label");
	}

	public CustomLabelRenderer(String field) {
		super(field);
		setHorizontalPadding(1);
	}

	/**
	 * @see prefuse.render.Renderer#render(java.awt.Graphics2D,
	 *      prefuse.visual.VisualItem)
	 */
	public void render(Graphics2D g, VisualItem item) {
		RectangularShape shape = (RectangularShape) getShape(g, item);
		if (shape == null)
			return;

		// fill the shape, if requested
		int type = getRenderType(item);
		if (type == RENDER_TYPE_FILL || type == RENDER_TYPE_DRAW_AND_FILL)
			GraphicsLib.paint(g, item, shape, getStroke(item), RENDER_TYPE_FILL);

		// now render the image and text
		String text = m_text;
		Image img = getImage(item);

		if (text == null && img == null)
			return;

		double size = item.getSize();
		boolean useInt = 1.5 > Math.max(g.getTransform().getScaleX(), g.getTransform().getScaleY());
		double x = shape.getMinX() + size * m_horizBorder;
		double y = shape.getMinY() + size * m_vertBorder;

		// render image
		if (img != null) {
			double w = size * img.getWidth(null);
			double h = size * img.getHeight(null);
			double ix = x, iy = y;

			// determine one co-ordinate based on the image position
			switch (m_imagePos) {
				case Constants.LEFT :
					x += w + size * m_imageMargin;
					break;
				case Constants.RIGHT :
					ix = shape.getMaxX() - size * m_horizBorder - w;
					break;
				case Constants.TOP :
					y += h + size * m_imageMargin;
					break;
				case Constants.BOTTOM :
					iy = shape.getMaxY() - size * m_vertBorder - h;
					break;
				default :
					throw new IllegalStateException("Unrecognized image alignment setting.");
			}

			// determine the other coordinate based on image alignment
			switch (m_imagePos) {
				case Constants.LEFT :
				case Constants.RIGHT :
					// need to set image y-coordinate
					switch (m_vImageAlign) {
						case Constants.TOP :
							break;
						case Constants.BOTTOM :
							iy = shape.getMaxY() - size * m_vertBorder - h;
							break;
						case Constants.CENTER :
							iy = shape.getCenterY() - h / 2;
							break;
					}
					break;
				case Constants.TOP :
				case Constants.BOTTOM :
					// need to set image x-coordinate
					switch (m_hImageAlign) {
						case Constants.LEFT :
							break;
						case Constants.RIGHT :
							ix = shape.getMaxX() - size * m_horizBorder - w;
							break;
						case Constants.CENTER :
							ix = shape.getCenterX() - w / 2;
							break;
					}
					break;
			}

			if (useInt && size == 1.0) {
				// if possible, use integer precision
				// results in faster, flicker-free image rendering
				g.drawImage(img, (int) ix, (int) iy, null);
			} else {
				m_transform.setTransform(size, 0, 0, size, ix, iy);
				g.drawImage(img, m_transform, null);
			}
		}

		// render text
		int textColor = item.getTextColor();
		if (text != null && ColorLib.alpha(textColor) > 0) {
			g.setPaint(ColorLib.getColor(textColor));
			g.setFont(m_font);
			FontMetrics fm = g.getFontMetrics(m_font);

			// compute available width
			double tw;
			switch (m_imagePos) {
				case Constants.TOP :
				case Constants.BOTTOM :
					tw = shape.getWidth() - 2 * size * m_horizBorder;
					break;
				default :
					tw = m_textDim.width;
			}

			// compute available height
			double th;
			switch (m_imagePos) {
				case Constants.LEFT :
				case Constants.RIGHT :
					th = shape.getHeight() - 2 * size * m_vertBorder;
					break;
				default :
					th = m_textDim.height;
			}

			// compute starting y-coordinate
			y += fm.getAscent();
			switch (m_vTextAlign) {
				case Constants.TOP :
					break;
				case Constants.BOTTOM :
					y += th - m_textDim.height;
					break;
				case Constants.CENTER :
					y += (th - m_textDim.height) / 2;
			}

			// render each line of text
			int lh = fm.getHeight(); // the line height
			int start = 0, end = text.indexOf(m_delim);
			for (; end >= 0; y += lh) {
				drawString(g, fm, text.substring(start, end), useInt, x, y, tw,item.getBoolean("temp"));
				start = end + 1;
				end = text.indexOf(m_delim, start);
			}
			drawString(g, fm, text.substring(start), useInt, x, y, tw,item.getBoolean("temp"));
		}

		// draw border
		if (type == RENDER_TYPE_DRAW || type == RENDER_TYPE_DRAW_AND_FILL) {
			GraphicsLib.paint(g, item, shape, getStroke(item), RENDER_TYPE_DRAW);
		}
	}

	public final void drawString(Graphics2D g, FontMetrics fm, String text, boolean useInt, double x, double y, double w, boolean temp) {
		// compute the x-coordinate
		if(temp)
			g.setColor(Color.CYAN);
		double tx;
		switch (m_hTextAlign) {
			case Constants.LEFT :
				tx = x;
				break;
			case Constants.RIGHT :
				tx = x + w - fm.stringWidth(text);
				break;
			case Constants.CENTER :
				tx = x + (w - fm.stringWidth(text)) / 2;
				break;
			default :
				throw new IllegalStateException("Unrecognized text alignment setting.");
		}
		// use integer precision unless zoomed-in
		// results in more stable drawing
		if (useInt) {
			g.drawString(text, (int) tx, (int) y);
		} else {
			g.drawString(text, (float) tx, (float) y);
		}
	}

	public Shape getShape(Graphics2D g, VisualItem item) {
		AffineTransform at = getTransform(item);
		Shape rawShape = getRawShape(g, item);
		return (at == null || rawShape == null ? rawShape : at.createTransformedShape(rawShape));
	}

	/**
	 * @see prefuse.render.AbstractShapeRenderer#getRawShape(prefuse.visual.VisualItem)
	 */
	protected Shape getRawShape(Graphics2D g, VisualItem item) {
		m_text = getText(item);
		Image img = getImage(item);
		double size = 1;

		// get image dimensions
		double iw = 0, ih = 0;
		if (img != null) {
			ih = img.getHeight(null);
			iw = img.getWidth(null);
		}

		// get text dimensions
		int tw = 0, th = 0;
		if (m_text != null) {
			if (item instanceof DecoratorItem) {
				setMaxTextWidth(Utils.getWidth(((DecoratorItem) item).getDecoratedItem())
						- (getHorizontalPadding() * 2));
			}
			m_text = computeTextDimensions(g, item, m_text, size);
			th = m_textDim.height;
			tw = m_textDim.width;
		}

		// get bounding box dimensions
		double w = 0, h = 0;
		switch (m_imagePos) {
			case Constants.LEFT :
			case Constants.RIGHT :
				w = tw + size * (iw + 2 * m_horizBorder + (tw > 0 && iw > 0 ? m_imageMargin : 0));
				h = Math.max(th, size * ih) + size * 2 * m_vertBorder;
				break;
			case Constants.TOP :
			case Constants.BOTTOM :
				w = Math.max(tw, size * iw) + size * 2 * m_horizBorder;
				h = th + size * (ih + 2 * m_vertBorder + (th > 0 && ih > 0 ? m_imageMargin : 0));
				break;
			default :
				throw new IllegalStateException("Unrecognized image alignment setting.");
		}

		// get the top-left point, using the current alignment settings
		getAlignedPoint(m_pt, item, w, h, m_xAlign, m_yAlign, g.getFont()); //PMD pass font as param

		if (m_bbox instanceof RoundRectangle2D) {
			RoundRectangle2D rr = (RoundRectangle2D) m_bbox;
			rr.setRoundRect(m_pt.getX(), m_pt.getY(), w, h, size * m_arcWidth, size * m_arcHeight);
		} else {
			m_bbox.setFrame(m_pt.getX(), m_pt.getY(), w, h);
		}
		return m_bbox;
	}

	protected String computeTextDimensions(Graphics2D g, VisualItem item, String text, double size) {
		// put item font in temp member variable
		m_font = item.getFont();
		// scale the font as needed
		if (size != 1) {
			m_font = FontLib.getFont(m_font.getName(), m_font.getStyle(), size * m_font.getSize());
		}
		FontMetrics dfm = DEFAULT_GRAPHICS.getFontMetrics(m_font);
		FontMetrics fm = g.getFontMetrics(m_font);
		StringBuffer str = null;

		//PMD if font size 8 (or less) then use specific settings
		m_font = FontLib.getFont(m_font.getName(),Font.BOLD,7);		//HACK use font size 3
		
		m_maxTextWidth = m_maxTextWidth*3;		//HACK PMD double
		boolean useDefault = true;
		if (fm.getFont().getSize() < 11) {
			useDefault = false;
		}

		// compute the number of lines and the maximum width
		int nlines = 1, w = 0, start = 0, end = text.indexOf(m_delim);
		m_textDim.width = 0;
		String line;
		for (; end >= 0; ++nlines) {
			w = fm.stringWidth(line = text.substring(start, end));
			// abbreviate line as needed
			if (m_maxTextWidth > -1 && w > m_maxTextWidth) {
				if (str == null)
					str = new StringBuffer(text.substring(0, start));
				//				str.append(abbreviateDots(line, fm, m_maxTextWidth));		//PMD changed to multiline
				str.append(multiLineLabel(line, fm, m_maxTextWidth,false));
				str.append(m_delim);
				w = m_maxTextWidth;
			} else if (str != null) {
				str.append(line).append(m_delim);
			}
			// update maximum width and substring indices
			m_textDim.width = Math.max(m_textDim.width, w);
			start = end + 1;
			end = text.indexOf(m_delim, start);
		}
		w = fm.stringWidth(line = text.substring(start));
		// abbreviate line as needed
		if (m_maxTextWidth > -1 && w >= m_maxTextWidth) {
			if (str == null)
				str = new StringBuffer(text.substring(0, start));
			//			str.append(abbreviateDots(line, dfm, m_maxTextWidth));		//PMD changed to multiline
			str.append(multiLineLabel(line, fm, m_maxTextWidth,false));
			w = m_maxTextWidth;
		} else if (str != null) {
			str.append(line);
		}
		// update maximum width
		m_textDim.width = Math.max(m_textDim.width, w);

		// compute the text height
		m_textDim.height = fm.getHeight() * nlines;

		return str == null ? text : str.toString();
	}

	private static String multiLineLabel(String str, FontMetrics fm, int width, boolean useDefault) {
		if (!useDefault) {
			int lastblank = 0, nchars = 0, cumx = 0;
			String str1 = "", str2 = "", str3 = ""; //3 line string
			//first line
			while (cumx < width && nchars < str.length()) {
				cumx += fm.charWidth(str.charAt(nchars));
				nchars++;
			}
			if (nchars < str.length()) { //first line
				str1 = str.substring(0, nchars - 1) + "-\n";
			}
			if (2 * nchars < str.length()) { //second line - check to see if there's also a third line 
				str2 = str.substring(nchars - 1, 2 * nchars - 1) + "-\n";
				str3 = str.substring(2 * nchars - 1, str.length());
				str3 = abbreviateDots(str3, fm, width);
			} else {
				str2 = str.substring(nchars - 1, str.length());
			}
			return str1 + str2 + str3;
		} else {
			return abbreviateDots(str, fm, width);
		}
	}

	public static String abbreviateDots(String str, FontMetrics fm, int width) {
		int lastblank = 0, nchars = 0, cumx = 0;
		while (cumx < width && nchars < str.length()) {
			if (Character.isWhitespace(str.charAt(nchars))) {
				lastblank = nchars;
			}
			cumx += fm.charWidth(str.charAt(nchars));
			nchars++;
		}
		if (nchars < str.length() && lastblank > 0) {
			nchars = lastblank;
		}

		if (nchars > 0) {
			str = "..." + str;
			str = StringLib.abbreviate(str, fm, width);
			return str.substring(3) + "...";
		}

//		if(nchars < str.length())
//			return str.substring(0, nchars -3) + "...";
		
		return str;
	}

	/**
	 * Helper method, which calculates the top-left co-ordinate of an item given
	 * the item's alignment.
	 * 
	 * @param font
	 */
	protected static void getAlignedPoint(Point2D p, VisualItem item, double w, double h, int xAlign, int yAlign,
			Font font) {

		double x = item.getX(), y = item.getY();
		if (Double.isNaN(x) || Double.isInfinite(x))
			x = 0; // safety check
		if (Double.isNaN(y) || Double.isInfinite(y))
			y = 0; // safety check

		if (xAlign == Constants.CENTER) {
			x = x - (w / 2);
		} else if (xAlign == Constants.RIGHT) {
			x = x - w;
		}
		if (yAlign == Constants.CENTER) {
			y = y - (h / 2);
		} else if (yAlign == Constants.BOTTOM) {
			y = y - h;
		}
		x = item.getX() - w / 2;
//		y = item.getY() - (h + 0.5 * h); //label placement
		p.setLocation(x, y);
	}
}
