package org.processmining.humbertogarciacaballero.models.model.layout.stage;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.EdgeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.NodeParams;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public abstract class LayoutStage {

	private HierarchicalLayoutContext context;

	public LayoutStage(HierarchicalLayoutContext context) {
		setContext(context);
	}

	public HierarchicalLayoutContext getContext() {
		return context;
	}

	public void setContext(HierarchicalLayoutContext context) {
		this.context = context;
	}

	public abstract int run(Graph graph);

	// Handy methods
	public static NodeParams getNodeParams(Node n) {
		return HierarchicalLayouter.getNodeParams(n);
	}

	public static EdgeParams getEdgeParams(Edge e) {
		return HierarchicalLayouter.getEdgeParams(e);
	}

	public static NodeType getNodeType(Node node) {
		return NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY));
	}

	public static boolean isDummy(Edge e) {
		return isDummy(e.getSourceNode()) || isDummy(e.getTargetNode());
	}

	public static boolean isDummy(Node n) {
		return getNodeParams(n).dummy;
	}

	public static boolean isReversed(Edge e) {
		return getEdgeParams(e).reversed;
	}

	public static boolean isTransition(Node node) {
		return NodeType.TRANSITION.equals(NodeType.valueOf(node.getString(PNConstants.NODE_TYPE_KEY)));
	}

	public static boolean isSimplified(Node node) {
		return getNodeParams(node).simplified;
	}

	public static boolean isSimplified(Edge edge) {
		return getEdgeParams(edge).simplified;
	}

	public static void unvisitNodes(Graph graph) {
		setVisitedNodes(graph, false);
	}

	public static void visitNodes(Graph graph) {
		setVisitedNodes(graph, true);
	}

	private static void setVisitedNodes(Graph graph, boolean visited) {
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			setVisitedNode(iter.next(), visited);
		}
	}

	public static void setVisitedNode(Node n, boolean visited) {
		((NodeParams) n.get(HierarchicalLayouter.NODE_PARAMS)).visited = visited;
	}

	public static boolean isVisited(Node n) {
		return ((NodeParams) n.get(HierarchicalLayouter.NODE_PARAMS)).visited;
	}

	/**
	 * Gets the root nodes of a graph. Root nodes are those nodes whose in
	 * degree is equals to 0. This method takes into account the reversed edges.
	 * 
	 * @param graph
	 *            Graph to extract the root nodes from.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static Iterator<Node> getRootNodes(Graph graph) {
		List<Node> result = new LinkedList<>();
		Iterator<Node> iter = graph.nodes();

		while (iter.hasNext()) {
			Node root = iter.next();

			if (root.getInDegree() == 0) {
				Iterator<Edge> iter2 = root.outEdges();

				boolean reversed = false;
				while (iter2.hasNext() && !reversed) {
					Edge e = iter2.next();
					if (((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
						reversed = true;
					}
				}

				if (!reversed)
					result.add(root);
			}
		}

		return result.iterator();
	}

	/**
	 * Gets the root nodes of a graph. Root nodes are those nodes whose in
	 * degree is equals to 0. This method takes into account the reversed edges.
	 * 
	 * @param graph
	 *            Graph to extract the root nodes from.
	 * @param nodeType
	 *            This type will be used to determine which nodes are returned.
	 * @return
	 */
	public static Iterator<Node> getRootNodes(Graph graph, NodeType nodeType) {
		List<Node> result = new LinkedList<>();

		Iterator<Node> iter = getRootNodes(graph);
		while (iter.hasNext()) {
			Node node = iter.next();
			if (getNodeType(node) == nodeType) {
				result.add(node);
			}
		}

		return result.iterator();
	}

	/**
	 * Gets the sink nodes of a graph. Sink nodes are those nodes whose out
	 * degree is equals to 0. This method takes into account the reversed
	 * dedges.
	 * 
	 * @param graph
	 *            Graph to extract the root nodes from.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public static Iterator<Node> getSinkNodes(Graph graph) {
		List<Node> result = new LinkedList<>();
		Iterator<Node> iter = graph.nodes();

		while (iter.hasNext()) {
			Node sink = iter.next();

			if (getRealOutDegree(sink) == 0) {
				result.add(sink);
			}
		}

		return result.iterator();
	}

	private static int getRealOutDegree(Node node) {
		int degree = 0;
		Iterator<Edge> iter = node.outEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (!((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
				degree++;
			}
		}

		iter = node.inEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
				degree++;
			}
		}

		return degree;
	}

	@SuppressWarnings("unchecked")
	private static Iterator<Node> getRealSuccessors(Node node) {
		List<Node> result = new LinkedList<>();

		// Outgoing edges
		Iterator<Edge> iter = node.outEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (!((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
				result.add(e.getTargetNode());
			}
		}

		// Incoming edges
		iter = node.inEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
				result.add(e.getSourceNode());
			}
		}

		return result.iterator();
	}

	@SuppressWarnings("unchecked")
	private static Iterator<Node> getRealSuccessors(Node node, NodeType nodeType) {
		List<Node> result = new LinkedList<>();

		// Outgoing edges
		Iterator<Edge> iter = node.outEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (!((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed
					&& getNodeType(e.getTargetNode()) == nodeType) {
				result.add(e.getTargetNode());
			}
		}

		// Incoming edges
		iter = node.inEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed
					&& getNodeType(e.getSourceNode()) == nodeType) {
				result.add(e.getSourceNode());
			}
		}

		return result.iterator();
	}

	@SuppressWarnings("unchecked")
	private static Iterator<Node> getNeighbors(Node node) {
		List<Node> result = new LinkedList<>();
		Iterator<Node> iter = node.neighbors();
		while (iter.hasNext()) {
			result.add(iter.next());
		}

		return result.iterator();
	}

	@SuppressWarnings("unchecked")
	private static Iterator<Node> getNeighbors(Node node, NodeType nodeType) {
		List<Node> result = new LinkedList<>();
		Iterator<Node> iter = node.neighbors();
		while (iter.hasNext()) {
			Node neighbor = iter.next();
			if (getNodeType(neighbor) == nodeType) {
				result.add(neighbor);
			}
		}

		return result.iterator();
	}

	public static Iterator<Node> getSuccessors(Node n, boolean undirected) {
		if (undirected) {
			return getNeighbors(n);
		}

		return getRealSuccessors(n);
	}

	public static Iterator<Node> getSuccessors(Node n, boolean undirected, NodeType nodeType) {
		if (undirected) {
			return getNeighbors(n, nodeType);
		}

		return getRealSuccessors(n, nodeType);
	}

	@SuppressWarnings("unchecked")
	public static Iterator<Node> getRealPredecessors(Node n) {
		List<Node> result = new LinkedList<>();

		// Incoming edges
		Iterator<Edge> iter = n.inEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (!((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
				result.add(e.getSourceNode());
			}
		}

		// Outgoing edges
		iter = n.outEdges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (((EdgeParams) e.get(HierarchicalLayouter.EDGE_PARAMS)).reversed) {
				result.add(e.getTargetNode());
			}
		}

		return result.iterator();
	}
}
