package org.processmining.humbertogarciacaballero.models.model.layout.stage.emplacer;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.EdgeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayerContainer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public class FastAndSimpleEmplacer extends Emplacer {
	
	private static final int LEFTMOST_UPPER = 0;
	private static final int RIGHTMOST_UPPER = 1;
	private static final int LEFTMOST_LOWER = 2;
	private static final int RIGHTMOST_LOWER = 3;
	
	public FastAndSimpleEmplacer(HierarchicalLayoutContext context) {
		super(context);
	}

	@SuppressWarnings("unchecked")
	@Override
	public Map<Node, Float> calculateCoordinates(Graph graph) {
		preprocessing(graph);
		
		Map<Node, Integer[]> result = new HashMap<>();
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node node = iter.next();
			result.put(node, new Integer[4]);
		}
		
		int ranges[][] = new int[4][];
		ranges[LEFTMOST_UPPER] = computeAssignment(graph, LEFTMOST_UPPER, result);
		ranges[RIGHTMOST_UPPER] = computeAssignment(graph, RIGHTMOST_UPPER, result);
		ranges[LEFTMOST_LOWER] = computeAssignment(graph, LEFTMOST_LOWER, result);
		ranges[RIGHTMOST_LOWER] = computeAssignment(graph, RIGHTMOST_LOWER, result);
		
		return balance(graph, ranges, result);
	}
	
	@SuppressWarnings("unchecked")
	private Map<Node, Float> balance(Graph graph, int[][] ranges, Map<Node, Integer[]> result) {
		int minWidth = Integer.MAX_VALUE;
		int minWidthAssignment = -1;
		if (ranges[LEFTMOST_UPPER][1] - ranges[LEFTMOST_UPPER][0] < minWidth) {
			minWidth = ranges[LEFTMOST_UPPER][1] - ranges[LEFTMOST_UPPER][0];
			minWidthAssignment = LEFTMOST_UPPER;
		}
		if (ranges[RIGHTMOST_UPPER][1] - ranges[RIGHTMOST_UPPER][0] < minWidth) {
			minWidth = ranges[RIGHTMOST_UPPER][1] - ranges[RIGHTMOST_UPPER][0];
			minWidthAssignment = RIGHTMOST_UPPER;
		}
		if (ranges[LEFTMOST_LOWER][1] - ranges[LEFTMOST_LOWER][0] < minWidth) {
			minWidth = ranges[LEFTMOST_LOWER][1] - ranges[LEFTMOST_LOWER][0];
			minWidthAssignment = LEFTMOST_LOWER;
		}
		if (ranges[RIGHTMOST_LOWER][1] - ranges[RIGHTMOST_LOWER][0] < minWidth) {
			minWidth = ranges[RIGHTMOST_LOWER][1] - ranges[RIGHTMOST_LOWER][0];
			minWidthAssignment = RIGHTMOST_LOWER;
		}
		shift(graph, LEFTMOST_UPPER, ranges[minWidthAssignment][0] - ranges[LEFTMOST_UPPER][0], result);
		shift(graph, LEFTMOST_LOWER, ranges[minWidthAssignment][0] - ranges[LEFTMOST_LOWER][0], result);
		shift(graph, RIGHTMOST_UPPER, ranges[minWidthAssignment][1] - ranges[RIGHTMOST_UPPER][1], result);
		shift(graph, RIGHTMOST_LOWER, ranges[minWidthAssignment][1] - ranges[RIGHTMOST_LOWER][1], result);

		Map<Node, Float> x = new HashMap<>();
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			Arrays.sort(result.get(v));
			x.put(v, (float) (result.get(v)[1] + result.get(v)[2]) / 2);
		}
		
		return x;
	}
	
	@SuppressWarnings("unchecked")
	private void shift(Graph graph, int assignment, int value, Map<Node, Integer[]> result) {
		if (value != 0) {
			Iterator<Node> iter = graph.nodes();
			while (iter.hasNext()) {
				Node v = iter.next();
				result.get(v)[assignment] += value;
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	private int[] computeAssignment(Graph graph, int mode, Map<Node, Integer[]> result) {
		Map<Node, Node> root = new HashMap<>();
		Map<Node, Node> align = new HashMap<>();
		Map<Node, Integer> blockWidths = new HashMap<>();
		verticalAligment(graph, root, align, mode);
		computeBlockWidths(graph, root, blockWidths);
		Map<Node, Integer> x = horizontalCompaction(graph, root, align, mode, blockWidths);
		
		int min = Integer.MAX_VALUE;
		int max = Integer.MIN_VALUE;
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			result.get(v)[mode] = x.get(v);
			if (x.get(v) > max) {
				max = x.get(v).intValue();
			}
			if (x.get(v) < min) {
				min = x.get(v).intValue();
			}
		}
		
		return new int[]{min, max};
	}
	
	private void preprocessing(Graph graph) {
		LayerContainer layerContainer = getContext().getLayerContainer();
		for (int i = 1; i <= layerContainer.getSize() - 2; i++) {
			int k0 = 0;
			int l = 0;
			List<Node> layerNodes = layerContainer.getNodes(i);
			List<Node> nextLayerNodes = layerContainer.getNodes(i+1);
			for (int l1 = 0; l1 < nextLayerNodes.size(); l1++) {
				Node v = nextLayerNodes.get(l1);
				boolean innersegment = false;
				int upperPosition = -1;
				if (isDummy(v)) {
					List<Node> upperNeighbors = layerContainer.getNeighbors(v, -1);
					for (Node upperNeigbor : upperNeighbors) {
						if (isDummy(upperNeigbor)) {
							innersegment = true;
							upperPosition = layerNodes.indexOf(upperNeigbor);
							break;
						}
					}
				}
				if (l1 == nextLayerNodes.size() - 1 || innersegment) {
					int k1 = layerNodes.size() - 1;
					if (innersegment) {
						k1 = upperPosition;
					}
					while (l <= l1) {
						List<Node> upperNeighbors = layerContainer.getNeighbors(nextLayerNodes.get(l), -1);
						for (int k = 0; k < layerNodes.size(); k++) {
							if (upperNeighbors.contains(layerNodes.get(k))) {
								if (k < k0 || k > k1) {
									Edge edgeToMark = graph.getEdge(layerNodes.get(k), nextLayerNodes.get(l));
									if (edgeToMark != null) {
										markType1Conflit(edgeToMark);
									}
									edgeToMark = graph.getEdge(nextLayerNodes.get(l), layerNodes.get(k));
									if (edgeToMark != null) {
										markType1Conflit(edgeToMark);
									}
								}
							}
						}
						l++;
					}
					k0 = k1;
				}
			}
		}
	}
	
	private void markType1Conflit(Edge e) {
		((EdgeParams)e.get(HierarchicalLayouter.EDGE_PARAMS)).type1Conflict = true;
	}
	
	private boolean isType1ConflictMarked(Edge e) {
		return ((EdgeParams)e.get(HierarchicalLayouter.EDGE_PARAMS)).type1Conflict;
	}
	
	@SuppressWarnings("unchecked")
	private void verticalAligment(Graph graph, Map<Node, Node> root, Map<Node, Node> align, int mode) {
		LayerContainer layerContainer = getContext().getLayerContainer();
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			root.put(v, v);
			align.put(v, v);
		}
		
		switch (mode) {
		case LEFTMOST_UPPER:
			for (int i = 1; i < layerContainer.getSize(); i++) {
				verticalAlignmentLeft(graph, root, align, layerContainer.getNodes(i), layerContainer.getNodes(i - 1), -1);
			}
			break;
		case RIGHTMOST_UPPER:
			for (int i = 1; i < layerContainer.getSize(); i++) {
				verticalAlignmentRight(graph, root, align, layerContainer.getNodes(i), layerContainer.getNodes(i - 1), -1);
			}
			break;
		case LEFTMOST_LOWER:
			for (int i = layerContainer.getSize() - 2; i >= 0; i--) {
				verticalAlignmentLeft(graph, root, align, layerContainer.getNodes(i), layerContainer.getNodes(i + 1), +1);
			}
			break;
		case RIGHTMOST_LOWER:
			for (int i = layerContainer.getSize() - 2; i >= 0; i--) {
				verticalAlignmentRight(graph, root, align, layerContainer.getNodes(i), layerContainer.getNodes(i + 1), +1);
			}
			break;
		}
	}
	
	private void verticalAlignmentLeft(Graph graph, Map<Node, Node> root, Map<Node, Node> align, List<Node> layerNodes, List<Node> nextLayerNodes, int nextLayer) {
		int r = Integer.MIN_VALUE;
		for (int k = 0; k < layerNodes.size(); k++) {
			List<Node> upperNeighbors = getContext().getLayerContainer().getNeighbors(layerNodes.get(k), nextLayer);
			if (upperNeighbors.size() > 0) {
				int d = upperNeighbors.size();
				for (int m = (int) Math.floor((d+1)/2f); m <= (int)Math.ceil((d+1)/2f); m++) {
					Node v = layerNodes.get(k);
					if (align.get(v).equals(v)) {
						Node u = upperNeighbors.get(m - 1);
						Edge e = graph.getEdge(u, v);
						if (e == null) e = graph.getEdge(v, u);
						if (e != null && !isType1ConflictMarked(e) && r < nextLayerNodes.indexOf(u)) {
							align.put(u, v);
							root.put(v, root.get(u));
							align.put(v, root.get(v));
							r = nextLayerNodes.indexOf(u);
						}
					}
				}
			}
		}
	}
	
	private void verticalAlignmentRight(Graph graph, Map<Node, Node> root, Map<Node, Node> align, List<Node> layerNodes, List<Node> nextLayerNodes, int nextLayer) {
		int r = Integer.MAX_VALUE;
		for (int k = layerNodes.size() - 1; k >= 0; k--) {
			List<Node> upperNeighbors = getContext().getLayerContainer().getNeighbors(layerNodes.get(k), nextLayer);
			if (upperNeighbors.size() > 0) {
				int d = upperNeighbors.size();
				for (int m = (int)Math.ceil((d+1)/2f); m >= (int) Math.floor((d+1)/2f); m--) {
					Node v = layerNodes.get(k);
					if (align.get(v).equals(v)) {
						Node u = upperNeighbors.get(m - 1);
						Edge e = graph.getEdge(u, v);
						if (e == null) e = graph.getEdge(v, u);
						if (e != null && !isType1ConflictMarked(e) && r > nextLayerNodes.indexOf(u)) {
							align.put(u, v);
							root.put(v, root.get(u));
							align.put(v, root.get(v));
							r = nextLayerNodes.indexOf(u);
						}
					}
				}
			}
		}
	}
	
	private Map<Node, Integer> horizontalCompaction(Graph graph, Map<Node, Node> root, Map<Node, Node> align, int mode, Map<Node, Integer> blockWidths) {
		switch (mode) {
		case LEFTMOST_LOWER:
		case LEFTMOST_UPPER:
			return horizontalCompactionLeft(graph, root, align, blockWidths);
		case RIGHTMOST_LOWER:
		case RIGHTMOST_UPPER:
		default:
			return horizontalCompactionRight(graph, root, align, blockWidths);
		}
	}
	
	
	@SuppressWarnings("unchecked")
	private Map<Node, Integer> horizontalCompactionLeft(Graph graph, Map<Node, Node> root, Map<Node, Node> align, Map<Node, Integer> blockWidths) {
		Map<Node, Node> sink = new HashMap<>();
		Map<Node, Integer> shift = new HashMap<>();
		Map<Node, Integer> x = new HashMap<>();
		
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node node = iter.next();
			sink.put(node, node);
			shift.put(node, Integer.MAX_VALUE);
			x.put(node, null);
		}
		
		// root coordinates relative to sink
		iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			if (root.get(v).equals(v)) {
				placeBlockLeft(v, root, align, sink, shift, x, blockWidths);
			}
		}
		
		// absolute coordinates
		iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			x.put(v, x.get(root.get(v)));
			if (shift.get(sink.get(root.get(v))) < Integer.MAX_VALUE) {
				x.put(v, x.get(v) + shift.get(sink.get(root.get(v))));
			}
		}
		
		return x;
	}
	
	@SuppressWarnings("unchecked")
	private Map<Node, Integer> horizontalCompactionRight(Graph graph, Map<Node, Node> root, Map<Node, Node> align, Map<Node, Integer> blockWidths) {
		Map<Node, Node> sink = new HashMap<>();
		Map<Node, Integer> shift = new HashMap<>();
		Map<Node, Integer> x = new HashMap<>();
		
		Iterator<Node> iter = graph.nodes();
		while (iter.hasNext()) {
			Node node = iter.next();
			sink.put(node, node);
			shift.put(node, Integer.MAX_VALUE);
			x.put(node, null);
		}
		
		// root coordinates relative to sink
		iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			if (root.get(v).equals(v)) {
				placeBlockRight(v, root, align, sink, shift, x, blockWidths);
			}
		}
		
		// absolute coordinates
		iter = graph.nodes();
		while (iter.hasNext()) {
			Node v = iter.next();
			x.put(v, x.get(root.get(v)));
			if (shift.get(sink.get(root.get(v))) < Integer.MAX_VALUE) {
				x.put(v, x.get(v) - shift.get(sink.get(root.get(v))));
			}
		}
		
		return x;
	}
	
	private void placeBlockLeft(Node v, Map<Node, Node> root, Map<Node, Node> align, Map<Node, Node> sink, Map<Node, Integer> shift, Map<Node, Integer> x, Map<Node, Integer> blockWidths) {
		if (x.get(v) == null) {
			x.put(v, 0);
			Node w = v;
			do {
				List<Node> layerNodes = getContext().getLayerContainer().getNodes(getContext().getLayerContainer().getLayer(w));
				int index = layerNodes.indexOf(w);
				if (index > 0) {
					Node u = root.get(layerNodes.get(index - 1));
					placeBlockLeft(u, root, align, sink, shift, x, blockWidths);
					if (sink.get(v).equals(v)) sink.put(v, sink.get(u));
					if (!sink.get(v).equals(sink.get(u))) {
						shift.put(sink.get(u), Math.min(shift.get(sink.get(u)), x.get(v) - x.get(u) - getContext().getIntracellSpacing() - (int)(0.5 * (blockWidths.get(u) + blockWidths.get(v)))));
					} else {
						x.put(v, Math.max(x.get(v), x.get(u) + getContext().getIntracellSpacing() + (int)(0.5 * (blockWidths.get(u) + blockWidths.get(v)))));
					}
				}
				w = align.get(w);
			} while(!w.equals(v));
		}
	}
	
	private void placeBlockRight(Node v, Map<Node, Node> root, Map<Node, Node> align, Map<Node, Node> sink, Map<Node, Integer> shift, Map<Node, Integer> x, Map<Node, Integer> blockWidths) {
		if (x.get(v) == null) {
			x.put(v, 0);
			Node w = v;
			do {
				List<Node> layerNodes = getContext().getLayerContainer().getNodes(getContext().getLayerContainer().getLayer(w));
				int index = layerNodes.indexOf(w);
				if (index < layerNodes.size() - 1) {
					Node u = root.get(layerNodes.get(index + 1));
					placeBlockRight(u, root, align, sink, shift, x, blockWidths);
					if (sink.get(v).equals(v)) sink.put(v, sink.get(u));
					if (!sink.get(v).equals(sink.get(u))) {
						shift.put(sink.get(u), Math.max(shift.get(sink.get(u)), x.get(u) - x.get(v) + getContext().getIntracellSpacing() + (int)(0.5 * (blockWidths.get(u) + blockWidths.get(v)))));
					} else {
						x.put(v, Math.min(x.get(v), x.get(u) - getContext().getIntracellSpacing() - (int)(0.5 * (blockWidths.get(u) + blockWidths.get(v)))));
					}
				}
				w = align.get(w);
			} while(!w.equals(v));
		}
	}
	
	private void computeBlockWidths(Graph graph, Map<Node, Node> root, Map<Node, Integer> blockWidths) {
	    Iterator<Node> iter = graph.nodes();
	    while (iter.hasNext()) {
	        blockWidths.put(iter.next(), 0);
	    }
	    iter = graph.nodes();
	    while (iter.hasNext()) {
	        Node v = iter.next();
	        Node r = root.get(v);
	        blockWidths.put(r, Math.max(blockWidths.get(r), xsize(v)));
	    }
	}
}
