package org.processmining.humbertogarciacaballero.models.model.layout.stage.sequencer;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import org.processmining.humbertogarciacaballero.models.model.layout.data.LayerContainer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.LayoutStage;

import prefuse.action.ActionList;
import prefuse.data.Edge;
import prefuse.data.Node;

public abstract class Sequencer extends LayoutStage {

    protected static final Logger s_logger = Logger.getLogger(ActionList.class.getName());

    public Sequencer(HierarchicalLayoutContext context) {
        super(context);
    }

    @SuppressWarnings("unchecked")
    protected int calculateCrossings(List<Node> layerNodes, List<Node> previousLayer, Node u, int uPosition, Node v,
            int vPosition) {
        int totalCrossings = 0;
        Iterator<Node> uNeighbors = u.neighbors();

        while (uNeighbors.hasNext()) {
            Node w = uNeighbors.next();
            if (previousLayer.contains(w)) {

                Iterator<Node> vNeighbors = v.neighbors();
                while (vNeighbors.hasNext()) {
                    Node z = vNeighbors.next();
                    if (previousLayer.contains(z)) {
                        int indexW = previousLayer.indexOf(w);
                        int indexZ = previousLayer.indexOf(z);
                        if ((uPosition < vPosition && indexW > indexZ) || (uPosition > vPosition && indexW < indexZ)) {
                            totalCrossings++;
                        }
                    }
                }
            }
            // else if (layerNodes.contains(w)) {
            // Iterator<Node> vNeighbors = v.neighbors();
            // while (vNeighbors.hasNext()) {
            // Node z = vNeighbors.next();
            // if (layerNodes.contains(z)) {
            // int indexW = previousLayer.indexOf(w);
            // int indexZ = previousLayer.indexOf(z);
            // int indexU = Math.min(uPosition, indexW);
            // indexW = Math.max(uPosition, indexW);
            // int indexV = Math.min(vPosition, indexZ);
            // indexZ = Math.max(vPosition, indexZ);
            //
            // if (indexU < indexV && indexV < indexW && (indexZ > indexW ||
            // indexZ < indexU)) {
            // totalCrossings++;
            // }
            // }
            // }
            // }
        }

        return totalCrossings;
    }

    protected float medianValue(LayerContainer order, Node node, int nextLayer) {
        List<Node> neighbors = order.getNeighbors(node, nextLayer);
        float[] medianValues = new float[neighbors.size()];
        int arrayCount = 0;

        for (Node neighbor : neighbors) {
            medianValues[arrayCount++] = order.getIndex(neighbor);
        }

        Arrays.sort(medianValues);

        if (arrayCount == 0) {
            return -1f;
        } else if (arrayCount % 2 == 1) {
            return medianValues[arrayCount / 2];
        } else if (arrayCount == 2) {
            return ((medianValues[0] + medianValues[1]) / 2.0f);
        } else {
            int medianPoint = arrayCount / 2;
            float leftMedian = medianValues[medianPoint - 1] - medianValues[0];
            float rightMedian = medianValues[arrayCount - 1] - medianValues[medianPoint];

            return (medianValues[medianPoint - 1] * rightMedian + medianValues[medianPoint] * leftMedian)
                    / (leftMedian + rightMedian);
        }
    }

    public float barycenterValue(LayerContainer order, Node node, int nextLayer) {
        float barycenter = 0;
        int elements = 0;
        List<Node> neighbors = order.getNeighbors(node, nextLayer);

        for (Node neighbor : neighbors) {
            barycenter += order.getIndex(neighbor);
            elements++;
        }

        return barycenter / elements;
    }

    public int crossing(final LayerContainer order, final int layer, int nextLayer) {
        List<Node> layerNodes = order.getNodes(layer);

        List<Edge> edges = new LinkedList<>();

        for (Node node : layerNodes) {
            Iterator<Edge> iter = node.edges();
            while (iter.hasNext()) {
                Edge edge = iter.next();
                Node neigbour = edge.getAdjacentNode(node);
                if (order.getLayer(neigbour) == layer + nextLayer) {
                    edges.add(edge);
                }
            }
        }

        Collections.sort(edges, new Comparator<Edge>() {
            @Override
            public int compare(Edge o1, Edge o2) {
                int nodeLayer1Rank;
                int nodeNextLayer1Rank;

                if (order.getLayer(o1.getSourceNode()) == layer) {
                    nodeLayer1Rank = getNodeParams(o1.getSourceNode()).rank;
                    nodeNextLayer1Rank = getNodeParams(o1.getTargetNode()).rank;
                } else {
                    nodeLayer1Rank = getNodeParams(o1.getTargetNode()).rank;
                    nodeNextLayer1Rank = getNodeParams(o1.getSourceNode()).rank;
                }

                int nodeLayer2Rank;
                int nodeNextLayer2Rank;

                if (order.getLayer(o2.getSourceNode()) == layer) {
                    nodeLayer2Rank = getNodeParams(o2.getSourceNode()).rank;
                    nodeNextLayer2Rank = getNodeParams(o2.getTargetNode()).rank;
                } else {
                    nodeLayer2Rank = getNodeParams(o2.getTargetNode()).rank;
                    nodeNextLayer2Rank = getNodeParams(o2.getSourceNode()).rank;
                }

                if (nodeLayer1Rank < nodeLayer2Rank)
                    return -1;
                if (nodeLayer1Rank > nodeLayer2Rank)
                    return 1;

                if (nodeNextLayer1Rank < nodeNextLayer2Rank)
                    return -1;
                if (nodeNextLayer1Rank > nodeNextLayer2Rank)
                    return 1;

                return 0;
            }
        });

        int[] positions = new int[edges.size()];
        int index = 0;
        for (Edge edge : edges) {
            if (order.getLayer(edge.getSourceNode()) == layer) {
                positions[index] = getNodeParams(edge.getTargetNode()).rank;
            } else {
                positions[index] = getNodeParams(edge.getSourceNode()).rank;
            }
            index++;
        }

        return insertionSort(positions); // Alternative version
//        return treeAccumulator(order.getSize(layer + nextLayer), positions); // Better performance
    }

    // Alternative implementation
    public int crossing(LayerContainer order) {
        int crossings = 0;

        for (int i = 0; i < order.getSize() - 1; i++) {
            crossings += crossing(order, i, +1);
        }

        return crossings;
    }

    private int treeAccumulator(int southlayerLength, int[] positions) {
        int firstIndex = 1;
        while (firstIndex < southlayerLength)
            firstIndex <<= 1;

        int treeSize = 2 * firstIndex - 1;
        firstIndex -= 1;
        int[] tree = new int[treeSize];
        Arrays.fill(tree, 0);

        int crossings = 0;
        for (Integer position : positions) {
            int index = position + firstIndex;
            tree[index]++;
            while (index > 0) {
                if (index % 2 == 1)
                    crossings += tree[index + 1];
                index = (index - 1) >> 1;
                tree[index]++;
            }
        }
        return crossings;
    }

    private int insertionSort(int[] arr) {
        int moves = 0;
        int i, j, newValue;
        for (i = 1; i < arr.length; i++) {
            newValue = arr[i];
            j = i;
            while (j > 0 && arr[j - 1] > newValue) {
                arr[j] = arr[j - 1];
                j--;
                moves++;
            }
            arr[j] = newValue;
        }
        return moves;
    }
}
