package org.processmining.humbertogarciacaballero.models.model.layout.stage.simplifier;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter;
import org.processmining.humbertogarciacaballero.models.model.layout.HierarchicalLayouter.NodeParams;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayerContainer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.LayoutStage;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.sequencer.Sequencer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.sequencer.SequencerException;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;

public class Simplifier extends LayoutStage {

	private Sequencer sequencer;

	public Simplifier(HierarchicalLayoutContext context, Sequencer sequencer) {
		super(context);
		this.sequencer = sequencer;
	}

	@Override
	public int run(Graph graph) {
		boolean shortLongModeActive = false;
		int improvement = 0;
		Set<Edge> edgesToSimplify = new HashSet<>();
		if (getContext().isSimplifySameLayer()) {
			simplifySameLayer(graph);
		}

		if (getContext().isSimplifyShortEdgesEnabled()) {
			simplifyShortEdges(graph, edgesToSimplify);
			shortLongModeActive = true;
		}

		if (getContext().isSimplifyLongEdgesEnabled()) {
			simplifyLongEdges(graph, edgesToSimplify);
			shortLongModeActive = true;
		}

		if (shortLongModeActive) {
			simplifyEdges(graph, edgesToSimplify);
			if (getContext().isReorderNodesEnabled()
					&& (!getContext().isSimplifyShortEdgesEnabled() || !getContext().isSimplifyLongEdgesEnabled())) {
				sequencer.run(graph);
			}
		} else {
			int iteration = 0;
			do {
				improvement = 0;
				for (int i = 0; i < getContext().getLayerContainer().getSize() - 1; i++) {
					improvement += simplifyLayer(graph, i, edgesToSimplify);
				}

				simplifyEdges(graph, edgesToSimplify);
				edgesToSimplify.clear();

				if (getContext().isReorderNodesEnabled()) {
					try {
						sequencer.run(graph);
					} catch (Exception e) {
						throw new SequencerException("There is a problem with the sequencer technique", e);
					}
				}

				iteration++;
			} while (improvement > getContext().getTotalCrossingThreshold()
					&& iteration < getContext().getSimplifierMaxLoops());
		}

		return 0;
	}

	@SuppressWarnings("unchecked")
	private int simplifyLayer(Graph graph, int layer, final Set<Edge> edgesToSimplify) {
		// Data structures
		LayerContainer layerContainer = getContext().getLayerContainer();
		int improvement = 0;
		final Map<Node, Map<Node, Integer>> crossingTransitions = new HashMap<>();

		int crossings = calculateCrossings(graph, layer, crossingTransitions);

		if (crossings > getContext().getCrossingThreshold()) {
			Comparator<Node> comparator = new Comparator<Node>() {
				@Override
				public int compare(Node o1, Node o2) {
					return Integer.compare(calculateSum(crossingTransitions.get(o1)),
							calculateSum(crossingTransitions.get(o2)));
				}
			};

			do {
				// Gets transition node with maximum crossings
				Node node = Collections.max(crossingTransitions.keySet(), comparator);
				int nodeLayer = layerContainer.getLayer(node);

				// Adds all edges for that transition node
				Iterator<Edge> iter = node.edges();
				while (iter.hasNext()) {
					Edge edge = iter.next();
					int adjacentLayer = layerContainer.getLayer(graph.getAdjacentNode(edge, node));

					if ((!isDummy(edge) || getContext().isSimplifyLongEdgesEnabled())
							&& (nodeLayer == layer && adjacentLayer == layer + 1
									|| adjacentLayer == layer && nodeLayer == layer + 1)) {
						edgesToSimplify.add(edge);
					}
				}

				// Decreases number of crossings and increases improvement.
				int amount = calculateSum(crossingTransitions.get(node));
				crossings -= amount;
				improvement += amount;

				// Updates collections
				for (Node crossing : crossingTransitions.get(node).keySet()) {
					crossingTransitions.get(crossing).remove(node);
				}
				crossingTransitions.remove(node);

				// Repeats the loop until crossings are less than threshold or
				// collection is empty
			} while (crossings >= getContext().getCrossingThreshold() && !crossingTransitions.isEmpty());
		}

		return improvement;
	}

	private int calculateSum(Map<Node, Integer> innerMap) {
		int sum = 0;
		for (Node node : innerMap.keySet()) {
			sum += innerMap.get(node);
		}

		return sum;
	}

	@SuppressWarnings("unchecked")
	private int calculateCrossings(Graph graph, int layer, Map<Node, Map<Node, Integer>> crossingTransitions) {
		LayerContainer layerContainer = getContext().getLayerContainer();
		List<Edge> edges = new ArrayList<Edge>();
		List<Node> layerNodes = getContext().getLayerContainer().getNodes(layer);
		int crossings = 0;

		for (Node node : layerNodes) {
			Iterator<Edge> iter = node.edges();
			while (iter.hasNext()) {
				Edge e = iter.next();
				if ((!isDummy(e) || getContext().isSimplifyLongEdgesEnabled())
						&& layerContainer.getLayer(graph.getAdjacentNode(e, node)) == layer + 1) {
					edges.add(e);
				}
			}
		}

		for (int i = 0; i < edges.size(); i++) {
			for (int j = i + 1; j < edges.size(); j++) {
				Edge e1 = edges.get(i);
				Edge e2 = edges.get(j);

				int uPosition, vPosition, wPosition, zPosition;

				Node u, v, w, z;

				if (layerContainer.getLayer(e1.getSourceNode()) == layer) {
					u = e1.getSourceNode();
					w = e1.getTargetNode();
				} else {
					u = e1.getTargetNode();
					w = e1.getSourceNode();
				}

				uPosition = layerContainer.getIndex(u);
				wPosition = layerContainer.getIndex(w);

				if (layerContainer.getLayer(e2.getSourceNode()) == layer) {
					v = e2.getSourceNode();
					z = e2.getTargetNode();
				} else {
					v = e2.getTargetNode();
					z = e2.getSourceNode();
				}

				vPosition = layerContainer.getIndex(v);
				zPosition = layerContainer.getIndex(z);

				if ((uPosition < vPosition && wPosition > zPosition)
						|| (uPosition > vPosition && wPosition < zPosition)) {
					// Crossing detected
					Node transition1, transition2;

					if (isTransition(u)) {
						transition1 = u;
					} else {
						transition1 = w;
					}

					if (isTransition(v)) {
						transition2 = v;
					} else {
						transition2 = z;
					}

					addCrossing(crossingTransitions, transition1, transition2);
					addCrossing(crossingTransitions, transition2, transition1);

					crossings++;
				}
			}
		}

		return crossings;
	}

	private void addCrossing(Map<Node, Map<Node, Integer>> map, Node node, Node crossing) {
		Map<Node, Integer> innerMap;
		if (map.containsKey(node)) {
			innerMap = map.get(node);
		} else {
			innerMap = new HashMap<>();
			map.put(node, innerMap);
		}

		if (innerMap.containsKey(crossing)) {
			innerMap.put(crossing, innerMap.get(crossing) + 1);
		} else {
			innerMap.put(crossing, 1);
		}
	}

	private void simplifyEdges(Graph graph, Set<Edge> edgesToSimplify) {
		for (Edge e : edgesToSimplify) {
			if (e.getBoolean(PNConstants.EDGE_SIMPLIFICABLE)) {
				Node source = e.getSourceNode();
				NodeParams sourceParams = HierarchicalLayouter.getNodeParams(source);
				Node target = e.getTargetNode();
				NodeParams targetParams = HierarchicalLayouter.getNodeParams(target);
				int fromLayer = getContext().getLayerContainer().getLayer(source),
						toLayer = getContext().getLayerContainer().getLayer(target);

				if (fromLayer < toLayer) {
					sourceParams.bottomSimplified = true;
					targetParams.topSimplified = true;
				} else if (fromLayer == toLayer) {
					sourceParams.bottomSimplified = true;
					targetParams.bottomSimplified = true;
				} else {
					sourceParams.topSimplified = true;
					targetParams.bottomSimplified = true;
				}

				sourceParams.simplified = true;
				targetParams.simplified = true;
				HierarchicalLayouter.getEdgeParams(e).simplified = true;
				getContext().addSimplificationParameter(source);
				getContext().addSimplificationParameter(target);
				getContext().addSimplificationParameter(getEdgeParams(e).clone());

				if (getContext().isReorderNodesEnabled()) {
					graph.removeEdge(e);
				}
			}
		}
	}

	@SuppressWarnings("unchecked")
	private int simplifySameLayer(Graph graph) {
		Iterator<Node> iter = graph.nodes();
		Set<Edge> set = new HashSet<>();
		while (iter.hasNext()) {
			Node node = iter.next();
			if (isTransition(node)) {
				Iterator<Edge> iter2 = node.edges();
				boolean simplify = false;

				// Looks for an edge in the same layer that node
				while (iter2.hasNext() && !simplify) {
					Edge edge = iter2.next();
					simplify = getContext().getLayerContainer().getDistance(edge) == 0;
				}

				// If it was found
				if (simplify) {
					int nodeLayer = getContext().getLayerContainer().getLayer(node);

					// Adds all edges for that transition node
					iter2 = node.edges();
					while (iter2.hasNext()) {
						Edge edge = iter2.next();
						int adjacentLayer = getContext().getLayerContainer()
								.getLayer(graph.getAdjacentNode(edge, node));

						// Simplifies all edges
						if (nodeLayer + 1 == adjacentLayer || nodeLayer == adjacentLayer) {
							set.add(edge);
						}
					}

					HierarchicalLayouter.getNodeParams(node).bottomSimplified = true;
				}
			}
		}

		simplifyEdges(graph, set);

		return set.size();
	}

	/**
	 * This method looks for edges with the attribute
	 * {@link PNConstants#EDGE_FORCE_SIMPLIFICATION} set to true.
	 */
	public void forceSimplifier(Graph graph) {
		Set<Edge> edgesToSimplify = new HashSet<>();
		Iterator<Edge> iter = graph.edges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (getEdgeParams(e).forceSimplification) {
				edgesToSimplify.add(e);
			}
		}

		simplifyEdges(graph, edgesToSimplify);
	}

	private int simplifyShortEdges(Graph graph, Set<Edge> edgesToSimplify) {
		int improvement = 0;
		Iterator<Edge> iter = graph.edges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (!isDummy(e)) {
				edgesToSimplify.add(e);
				improvement++;
			}
		}
		return improvement;
	}

	/**
	 * Auxuliar method that simplifies long edges.
	 * 
	 * @param graph
	 */
	private int simplifyLongEdges(Graph graph, Set<Edge> edgesToSimplify) {
		int improvement = 0;
		Iterator<Edge> iter = graph.edges();
		while (iter.hasNext()) {
			Edge e = iter.next();
			if (isDummy(e)) {
				edgesToSimplify.add(e);
				improvement++;
			}
		}
		return improvement;
	}

	public void setSequencer(Sequencer sequencer) {
		this.sequencer = sequencer;
	}
}
