package org.processmining.humbertogarciacaballero.models.util;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.model.NodeType;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;
import org.processmining.humbertogarciacaballero.models.util.circuits.ElementaryCyclesSearch;
import org.processmining.models.graphbased.directed.transitionsystem.CoverabilityGraph;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Schema;
import prefuse.data.Tuple;
import prefuse.data.expression.Predicate;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.util.ColorLib;
import prefuse.util.PrefuseLib;
import prefuse.visual.AggregateItem;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualGraph;
import prefuse.visual.VisualItem;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;

public class Utils {

    // Prevents instantiation.
    private Utils() {
    }
    private static final Predicate ROOT_NODES_PREDICATE = ExpressionParser.predicate("INDEGREE() = 0");
    private static final Predicate SINK_NODES_PREDICATE = ExpressionParser.predicate("OUTDEGREE() = 0");
    private static final Predicate TOP_SIMPLIFIED_PREDICATE = ExpressionParser
            .predicate("[" + PNConstants.SIMPLIFIED_TOP_KEY + "] = TRUE AND [" + VisualItem.VISIBLE + "] = TRUE");
    private static final Predicate BOTTOM_SIMPLIFIED_PREDICATE = ExpressionParser
            .predicate("[" + PNConstants.SIMPLIFIED_BOTTOM_KEY + "] = TRUE AND [" + VisualItem.VISIBLE + "] = TRUE");
    private static final Predicate PLACE_TYPE_PREDICATE = ExpressionParser
            .predicate("[" + PNConstants.NODE_TYPE_KEY + "] = '" + NodeType.PLACE + "'");
    private static final Predicate TRANSITION_TYPE_PREDICATE = ExpressionParser
            .predicate("[" + PNConstants.NODE_TYPE_KEY + "] = '" + NodeType.TRANSITION + "'");

    private static final StringBuilder builder = new StringBuilder();
    private static final Schema NODE_SCHEMA = new Schema();

    static {
        NODE_SCHEMA.addColumn(PNConstants.NODE_ID_KEY, String.class, null);
        NODE_SCHEMA.addColumn(PNConstants.NODE_TYPE_KEY, String.class, NodeType.PLACE.toString());
        NODE_SCHEMA.addColumn(PNConstants.NODE_LABEL_KEY, String.class, PNConstants.EMPTY_LABEL);
        NODE_SCHEMA.addColumn(PNConstants.NODE_WIDTH_KEY, int.class, PNConstants.PLACES_DEFAULT_SIZE);
        NODE_SCHEMA.addColumn(PNConstants.NODE_HEIGHT_KEY, int.class, PNConstants.PLACES_DEFAULT_SIZE);
        NODE_SCHEMA.addColumn(PNConstants.VISIBLE_IN_DEGREE, int.class, -1);
        NODE_SCHEMA.addColumn(PNConstants.VISIBLE_OUT_DEGREE, int.class, -1);
        NODE_SCHEMA.addColumn(PNConstants.NODE_TAU_TRANSITION_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_LAYER_KEY, int.class, 0);
        NODE_SCHEMA.addColumn(PNConstants.NODE_DEPTH_KEY, int.class, 0);
        NODE_SCHEMA.addColumn(PNConstants.SIMPLIFIED_TOP_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.SIMPLIFIED_BOTTOM_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_PARALLEL_NODES_KEY, List.class, null);
        NODE_SCHEMA.addColumn(PNConstants.NODE_CHOICE_NODES_KEY, List.class, null);
        NODE_SCHEMA.addColumn(PNConstants.NODE_STARTING_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_ENDING_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_TOKEN_COUNT_KEY, int.class, 0);
        NODE_SCHEMA.addColumn(PNConstants.SIMPLIFIED_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.LOOP_KEY, int.class, -1);
        NODE_SCHEMA.addColumn(PNConstants.FILLING_EXTRA_COLOR_KEY, int.class, 0);
        NODE_SCHEMA.addColumn(PNConstants.NODE_NON_LIVE_TRANSITION_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_DEAD_TRANSITION_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_UNBOUNDED_PLACE_KEY, boolean.class, false);
        NODE_SCHEMA.addColumn(PNConstants.NODE_SELECTED_TRANSITION, boolean.class, false);
        NODE_SCHEMA.addColumn("placeLogMove", boolean.class, false);
        NODE_SCHEMA.lockSchema();
    }

    private static final Schema EDGE_SCHEMA = new Schema();

    static {
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_BENDS_KEY, List.class, null);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_SIMPLIFICABLE, boolean.class, true);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_FORCE_SIMPLIFICATION, boolean.class, false);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_SOURCE_PORT_OFFSET, float.class, 0f);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_TARGET_PORT_OFFSET, float.class, 0f);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_COLOR_KEY, int.class, PNConstants.EDGE_DEFAULT_COLOR);
        EDGE_SCHEMA.addColumn(PrefuseLib.getStartField(PNConstants.EDGE_COLOR_KEY), int.class,
                PNConstants.EDGE_DEFAULT_COLOR);
        EDGE_SCHEMA.addColumn(PrefuseLib.getEndField(PNConstants.EDGE_COLOR_KEY), int.class,
                PNConstants.EDGE_DEFAULT_COLOR);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_BIDIRECTIONAL, boolean.class, false);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_BIDIRECTIONAL_DRAW, boolean.class, true);
        EDGE_SCHEMA.addColumn(PNConstants.SIMPLIFIED_KEY, boolean.class, false);
        EDGE_SCHEMA.addColumn(PNConstants.SIMPLIFIED_OLD_KEY, boolean.class, false);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_RESET_KEY, boolean.class, false);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_INHIBITOR_KEY, boolean.class, false);
        EDGE_SCHEMA.addColumn(PNConstants.LOOP_KEY, int.class, -1);
        EDGE_SCHEMA.addColumn(PNConstants.EDGE_LONG_EDGE_KEY, boolean.class, false);
        EDGE_SCHEMA.lockSchema();
    }

    public static Rectangle2D getBounds(AggregateItem item) {
        @SuppressWarnings("unchecked")
        Iterator<VisualItem> iter = item.items();

        Rectangle2D bounds = new Rectangle2D.Double();
        if (iter.hasNext()) {
            bounds = iter.next().getBounds();
            while (iter.hasNext()) {
                bounds = bounds.createUnion(iter.next().getBounds());
            }
        }

        return bounds;
    }

    public static int getWidth(Tuple n) {
    	if (n.getBoolean(PNConstants.NODE_TAU_TRANSITION_KEY)) {
    		return PNConstants.TAU_TRANSITIONS_SIZE;
    	}
        return n.getInt(PNConstants.NODE_WIDTH_KEY);
    }

    public static int getHeight(Tuple n) {
        return n.getInt(PNConstants.NODE_HEIGHT_KEY);
    }

    public static void setPosition(VisualItem item, Point2D position) {
        PrefuseLib.setX(item, item, position.getX());
        PrefuseLib.setY(item, item, position.getY());
    }

    public static Rectangle2D getBounds(VisualGraph graph) {
        return graph.getVisualization().getBounds(graph.getGroup());
    }

    public static void initSchema(Graph graph) {
        Schema nodeSchema = graph.getNodeTable().getSchema();
        for (int i = 0; i < NODE_SCHEMA.getColumnCount(); i++) {
            if (nodeSchema.getColumnIndex(NODE_SCHEMA.getColumnName(i)) == -1) {
                graph.getNodes().addColumn(NODE_SCHEMA.getColumnName(i), NODE_SCHEMA.getColumnType(i),
                        NODE_SCHEMA.getDefault(i));
            }
        }

        Schema edgeSchema = graph.getEdgeTable().getSchema();
        for (int i = 0; i < EDGE_SCHEMA.getColumnCount(); i++) {
            if (edgeSchema.getColumnIndex(EDGE_SCHEMA.getColumnName(i)) == -1) {
                graph.getEdges().addColumn(EDGE_SCHEMA.getColumnName(i), EDGE_SCHEMA.getColumnType(i),
                        EDGE_SCHEMA.getDefault(i));
            }
        }
        
        //PMD starts
        graph.addColumn("ID", int.class);
		graph.getNodes().addColumn("Preceding", String.class);
		graph.getNodes().addColumn("Follows", String.class);
		graph.getNodes().addColumn("CoOccurs", String.class);
		graph.getNodes().addColumn("ChoicePlaces1", ArrayList.class);
		graph.getNodes().addColumn("InputOutputNodes", ArrayList.class);
		graph.getNodes().addColumn("LoopPlaces", ArrayList.class);
		graph.getNodes().addColumn("PrePostNodes", ArrayList.class);
		graph.getNodes().addColumn("x", double.class);
		graph.getNodes().addColumn("y", double.class);
		graph.getEdges().addColumn("temp", boolean.class); //these are temporarily added edges - used diff color
		graph.getNodes().addColumn("temp", boolean.class);		//these are temporarily added nodes - used diff color
		//PMD ends
    }

    @SuppressWarnings("unchecked")
    public static int markLoops(Graph graph) {
        ElementaryCyclesSearch el = new ElementaryCyclesSearch(getAdjacencyMatrix(graph),
                Lists.newArrayList(graph.nodes()).toArray());
        List<List<Node>> cycles = el.getElementaryCycles();
        int loop = 0;
        for (List<Node> cycle : cycles) {
            Set<Node> set = Sets.newHashSet(cycle);
            for (Node node : cycle) {
                Iterator<Edge> iter = node.edges();
                while (iter.hasNext()) {
                    Edge e = iter.next();
                    if (set.contains(e.getAdjacentNode(node)))
                        e.setInt(PNConstants.LOOP_KEY, loop);
                }
            }
            loop++;
        }
        return loop;
    }
    
    public static boolean[][] getAdjacencyMatrix(Graph g) {
        boolean[][] adjacencyMatrix = new boolean[g.getNodeCount()][g.getNodeCount()];

        Iterator<Node> iter = g.nodes();
        while (iter.hasNext()) {
            Node node = iter.next();
            Iterator<Node> edges = node.outNeighbors();
            while (edges.hasNext()) {
                Node adjacent = edges.next();
                adjacencyMatrix[node.getRow()][adjacent.getRow()] = true;
            }
        }

        return adjacencyMatrix;
    }

    public static String toName(int number) {
        builder.setLength(0);
        while (number-- > 0) {
            builder.append((char) ('A' + (number % 26)));
            number /= 26;
        }
        return builder.reverse().toString();
    }

    public static Point2D calculateOffsetPortPoint(EdgeItem edgeItem, NodeItem nodeItem, boolean source,
            boolean downward) {
        Point2D port = calculatePortPoint(edgeItem, nodeItem, source);
        float landingOffset = HierarchicalLayoutContext.getInstance().getEdgeLandingOffset();
        double size;
        switch (HierarchicalLayoutContext.getInstance().getLayoutOrientation()) {
        case TOP_TO_BOTTOM:
            size = Utils.getHeight(nodeItem) / 2f;
            return new Point2D.Double(port.getX(),
                    port.getY() + (downward ? landingOffset + size : -landingOffset - size));
        case BOTTOM_TO_TOP:
            size = Utils.getHeight(nodeItem) / 2f;
            return new Point2D.Double(port.getX(),
                    port.getY() + (downward ? -landingOffset - size : landingOffset + size));
        case LEFT_TO_RIGHT:
            size = Utils.getWidth(nodeItem) / 2f;
            return new Point2D.Double(port.getX() + (downward ? landingOffset + size : -landingOffset - size),
                    port.getY());
        default:
            size = Utils.getWidth(nodeItem) / 2f;
            return new Point2D.Double(port.getX() + (downward ? -landingOffset - size : landingOffset + size),
                    port.getY());
        }
    }

    public static Point2D calculatePortPoint(EdgeItem edgeItem, NodeItem nodeItem, boolean source) {
        float offset = edgeItem
                .getFloat(source ? PNConstants.EDGE_SOURCE_PORT_OFFSET : PNConstants.EDGE_TARGET_PORT_OFFSET);
        switch (HierarchicalLayoutContext.getInstance().getLayoutOrientation()) {
        case TOP_TO_BOTTOM:
            return new Point2D.Double(nodeItem.getX() - offset, nodeItem.getY());
        case BOTTOM_TO_TOP:
            return new Point2D.Double(nodeItem.getX() + offset, nodeItem.getY());
        case LEFT_TO_RIGHT:
            return new Point2D.Double(nodeItem.getX(), nodeItem.getY() + offset);
        default:
            return new Point2D.Double(nodeItem.getX(), nodeItem.getY() - offset);
        }
    }

    public static boolean checkTopSimplification(AggregateItem item) {
        @SuppressWarnings("unchecked")
        Iterator<VisualItem> iter = item.items(TOP_SIMPLIFIED_PREDICATE);
        return iter.hasNext();
    }

    public static boolean checkBottomSimplification(AggregateItem item) {
        @SuppressWarnings("unchecked")
        Iterator<VisualItem> iter = item.items(BOTTOM_SIMPLIFIED_PREDICATE);
        return iter.hasNext();
    }

    public static void higlightPathFromBeginning(Graph g, NodeItem item) {
    	CoverabilityGraph cg = PNetVisualizerController.getInstance().getCoverabilityGraph();
    	
    }
    
    public static int changeAlphaDepth(int color, int depth) {
    	return ColorLib.setAlpha(color, Math.max(0, 255 / depth));
    }
    
    /**
     * Returns an iterator over the root nodes of the specified graph. The
     * returned iterator contains tuples instead of nodes.
     * 
     * @param g
     *            Graph to look the root nodes for.
     * @return
     */
    @SuppressWarnings("unchecked")
    public static Iterator<Node> getRootNodes(Graph g) {
        List<Node> list = new LinkedList<>();
        Iterator<Node> iter = g.nodes();
        while (iter.hasNext()) {
            Node node = iter.next();
            if (ROOT_NODES_PREDICATE.getBoolean(node)) {
                list.add(node);
            }
        }

        return list.iterator();
    }

    /**
     * Returns an iterator over the sink nodes of the specified graph. The
     * returned iterator contains tuples instead of nodes.
     * 
     * @param g
     *            Graph to look the sink nodes for.
     * @return
     */
    @SuppressWarnings("unchecked")
    public static Iterator<Node> getSinkNodes(Graph g) {
        List<Node> list = new LinkedList<>();
        Iterator<Node> iter = g.nodes();
        while (iter.hasNext()) {
            Node node = iter.next();
            if (SINK_NODES_PREDICATE.getBoolean(node)) {
                list.add(node);
            }
        }

        return list.iterator();
    }
}
